﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Transactions;
using HeuristicLab.Services.Hive.Common;
using HeuristicLab.Services.Hive.Common.DataTransfer;
using HeuristicLab.Tracing;
using HeuristicLab.Core;

namespace HeuristicLab.Services.Hive {
  /// <summary>
  /// This class holds the state of all recent heartbeats and decides to reschedule jobs and set slaves offline
  /// </summary>
  public class LifecycleManager : ILifecycleManager {
    private DataAccess.IHiveDao dao {
      get { return ServiceLocator.Instance.HiveDao; }
    }
    private HeuristicLab.Services.Hive.DataAccess.TransactionManager trans {
      get { return ServiceLocator.Instance.TransactionManager; }
    }
    private IAuthorizationManager auth {
      get { return ServiceLocator.Instance.AuthorizationManager; }
    }

    private static object locker = new object();

    // Windows-Forms timer is single threaded, so callbacks will be synchron
    System.Windows.Forms.Timer timer;

    public ExecutionState ExecutionState {
      get { return timer.Enabled ? ExecutionState.Started : Core.ExecutionState.Stopped; }
    }

    public LifecycleManager() {
      this.timer = new System.Windows.Forms.Timer();
      this.timer.Tick += new EventHandler(timer_Tick);
    }

    public void Start() {
      if (ExecutionState == Core.ExecutionState.Stopped) {
        this.timer.Interval = (int)new TimeSpan(0, 0, 10).TotalMilliseconds;
        this.timer.Start();
      }
    }

    public void Stop() {
      if (ExecutionState == Core.ExecutionState.Started) {
        timer.Stop();
      }
    }

    /// <summary>
    /// This method is supposed to check if slaves are online
    /// if not -> set them offline and check if they where calculating a job
    /// </summary>
    private void timer_Tick(object sender, EventArgs e) {
      lock (locker) {
        using (trans.OpenTransaction()) {
          var slaves = dao.GetSlaves(x => x.SlaveState != SlaveState.Offline);
          foreach (Slave slave in slaves) {
            if (!slave.LastHeartbeat.HasValue || (DateTime.Now - slave.LastHeartbeat.Value).TotalSeconds > ApplicationConstants.HeartbeatTimeout) {
              slave.SlaveState = SlaveState.Offline;
              AbortJobs(slave.Id);
              dao.UpdateSlave(slave);
            }
          }
        }
      }
    }

    private void AbortJobs(Guid slaveId) {
      var jobs = dao.GetJobs(x => x.Slave.ResourceId == slaveId);
      foreach (var j in jobs) {
        j.JobState = JobState.Waiting;
        dao.UpdateJob(j);
      }
    }
  }
}
